import pytest
from pyautodiff.dual import Dual
import pyautodiff.admath as admath
import numpy as np
import math


def test1__sin_fn__():
    x = Dual(0)
    f = admath.sin(x)
    assert f.val == 0
    assert f.der == 1


def test2__sin_fn__():
    tests = np.linspace(-10,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.sin(x)
        assert np.sin(test) == admath.sin(test)
        assert np.sin(test) == f.val
        assert np.cos(test) == f.der

def test1__cos_fn__():
    x = Dual(0)
    f = admath.cos(x)
    assert f.val == np.cos(0)
    assert f.der == -1*np.sin(0)


def test2__cos_fn__():
    tests = np.linspace(-10,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.cos(x)
        assert np.cos(test) == admath.cos(test)
        assert np.cos(test) == f.val
        assert -1*np.sin(test) == f.der

def test1__tan_fn__():
    x = Dual(0)
    f = admath.tan(x)
    assert f.val == 0
    assert f.der == 1
    assert admath.tan(0) == 0


def test_log_fn__():
    tests = np.linspace(1,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.log(x)
        assert np.log(test) == admath.log(test)
        assert np.log(test) == f.val
        assert 1/test == f.der

def test_log2_fn__():
    tests = np.linspace(1,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.log2(x)
        assert np.log2(test) == admath.log2(test)
        assert np.log2(test) == f.val
        assert 1/(test*np.log(2)) == f.der

def test_log10_fn__():
    tests = np.linspace(1,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.log10(x)
        assert np.log10(test) == admath.log10(test)
        assert np.log10(test) == f.val
        assert 1/(test*np.log(10)) == f.der

def test_logb__():
    tests = np.linspace(1,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.logb(x, np.exp(1))
        assert np.log(test) == admath.logb(test, np.exp(1))
        assert np.log(test) == f.val
        assert 1/test == f.der

def test_exp_fn__():
    tests = np.linspace(-100,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.exp(x)
        assert np.exp(test) == admath.exp(test)
        assert np.exp(test) == f.val
        assert np.exp(test) == f.der

def test_power1():
    x = Dual(2)
    f = admath.power(x,x)
    assert f.der == 2*(2+2*np.log(2))
    assert f.val == 4.0
    assert admath.power(2,2) == 4.0
    f = admath.power(x,2)
    assert f.val == 4
    assert f.der == 4
    f = admath.power(2,x)
    assert f.val == 4.0
    assert f.der == np.log(2) * 2 ** 2

def test_sqrt_fn__():
    tests = np.linspace(1,100,15)
    for test in tests:
        x = Dual(test)
        f = admath.sqrt(x)
        assert np.sqrt(test) == admath.sqrt(test)
        assert np.sqrt(test) == f.val
        assert 1/2*1/np.sqrt(test) == f.der

def test_arcsin():
    x = Dual(0)
    f = admath.arcsin(x)
    assert 1 == f.der
    assert 0 == f.val
    assert 0 == admath.arcsin(0)

def test_arccos():
    x = Dual(0)
    f = admath.arccos(x)
    assert -1 == f.der
    assert np.pi/2 == f.val
    assert np.pi/2 == admath.arccos(0)

def test_arctan():
    x = Dual(0)
    f = admath.arctan(x)
    assert 1 == f.der
    assert 0 == f.val
    assert 0 == admath.arctan(0)

def test_sinh():
    x = Dual(0)
    f = admath.sinh(x)
    assert 1 == f.der
    assert 0 == f.val
    assert 0 == admath.sinh(0)

def test_cosh():
    x = Dual(0)
    f = admath.cosh(x)
    assert 0 == f.der
    assert 1 == f.val
    assert 1 == admath.cosh(0)

def test_tanh():
    x = Dual(0)
    f = admath.tanh(x)
    assert 1 == f.der
    assert 0 == f.val
    assert 0 == admath.tanh(0)

def test_logistic():
    x = Dual(1)
    f = admath.logistic(x)
    assert 0.25 == f.der
    assert 0.50 == f.val
    assert 0.50 == admath.logistic(1)

def test_logistic():
    x = Dual(1)
    f = admath.logistic(x)
    assert 0.25 == f.der
    assert 0.50 == f.val
    assert 0.50 == admath.logistic(1)

def test_sum():
    x = Dual(1)
    f = admath.sum([x**2, x**3])
    assert 5 == f.der
    assert 2 == f.val
    assert 5 == admath.sum([2,3])

def test_abs():
    x = Dual(-1)
    f = admath.abs(x**3)
    assert -1 == f.der
    assert 1 == f.val
    assert 1 == admath.abs(-1)



import pytest
from pyautodiff.dual import Dual
import numpy as np

## Unary operation test
def test__neg__():
    x = Dual(3)
    f = -x
    assert f.val == -3
    assert f.der == -1

def test__pos__():
    x = Dual(-3)
    f = +x
    assert f.val == -3
    assert f.der == 1


## Addition operation test

def test1__add__():
    x = Dual(3)
    f = 6*x + 9
    assert f.val == 27
    assert f.der == 6

def test2__add__():
    x = Dual(2)
    f = 6*x + 4*x
    assert f.val == 20
    assert f.der == 10

def test__radd__():
    x = Dual(2)
    f = 4 + 4*x
    assert f.val == 12
    assert f.der == 4



## Substraction operation test

def test1__sub__():
    x = Dual(2)
    f = 6*x - 9
    assert f.val == 3
    assert f.der == 6

def test2__sub__():
    x = Dual(3)
    f = 6*x - 4*x
    assert f.val == 6
    assert f.der == 2

def test__rsub__():
    x = Dual(1)
    f = 4 - 4*x
    assert f.val == 0
    assert f.der == -4



## Multiplication operation test

def test1__mul__():
    x = Dual(2)
    f = x * 9
    assert f.val == 18
    assert f.der == 9

def test2__mul__():
    x = Dual(1)
    f = x * x
    assert f.val == 1
    assert f.der == 2

def test__rmul__():
    x = Dual(3)
    f = 4 * x
    assert f.val == 12
    assert f.der == 4



## Division operation test

def test1__truediv__():
    x = Dual(2)
    f = x / 10
    assert f.val == 0.2
    assert f.der == 0.1

def test2__truediv__():
    x = Dual(1)
    f = (2*x ** 2) / (4*x)
    assert f.val == 0.5
    assert f.der == 0.5

def test__rtruediv__():
    x = Dual(2)
    f =  1 / x
    assert f.val == 0.5
    assert f.der == -0.25



# Power operation test

def test1__pow__():
    x = Dual(2)
    f = x ** x
    assert f.val == 4.0
    assert f.der == 2*(2+2*np.log(2))

def test2__pow__():
    x = Dual(2)
    f = x ** 2
    assert f.val == 4
    assert f.der == 4

def test__rpow__():
    x = Dual(2)
    f = 2 ** x
    assert f.val == 4.0
    assert f.der == np.log(2) * 2 ** 2

# Comparision operations equal and not equal

def test__eq__():
    x = Dual(2)
    y = Dual(2)
    z = Dual(1)
    assert True == (x == y)
    assert False == (x == z)
    assert True == (x == 2)
    assert True == (2 == x)
    assert False == (x == 3)
    assert False == (3 == x)

def test_ne__():
    x = Dual(2)
    y = Dual(2)
    z = Dual(1)
    assert False == (x != y)
    assert True == (x != z)
    assert True == (x != 4)
    assert True == (4 != x)
    assert False == (x != 2)
    assert False == (2 != x)

# Comparision operation

def test__lt__():
    x = Dual(2)
    y = Dual(2)
    z = Dual(1)
    assert False == (x < y)
    assert False == (x < z)
    assert True  == (z < x)
    assert True  == (x < 3)
    assert False == (3 < x)


def test__gt__():
    x = Dual(2)
    y = Dual(2)
    z = Dual(1)
    assert False == (x > y)
    assert True  == (x > z)
    assert False == (z > x)
    assert True  == (x > 1)
    assert False == (1 > x)

def test__le__():
    x = Dual(2)
    y = Dual(2)
    z = Dual(1)
    assert True  == (x <= y)
    assert False == (x <= z)
    assert True  == (z <= x)
    assert True  == (x <= 3)
    assert False == (3 <= x)

def test__ge__():
    x = Dual(2)
    y = Dual(2)
    z = Dual(1)
    assert True  == (x >= y)
    assert True  == (x >= z)
    assert False == (z >= x)
    assert True  == (x >= 1)
    assert False == (1 >= x)



import pytest
from pyautodiff.interface import AutoDiff as AD
import pyautodiff.admath as admath

import numpy as np


def my_fn_1d(x,y):
	return x**2 + y**2

def test_list():
	fn = AD(my_fn_1d)
	der = fn.get_der([1,2])
	val = fn.get_val([1,2])
	assert der == [2, 4]
	assert val == 5

def test_list_lists():
	fn = AD(my_fn_1d)
	der = fn.get_der([[1,2],[3,4],[5,6]])
	val = fn.get_val([[1,2],[3,4],[5,6]])
	assert der == [[2, 4], [6, 8], [10, 12]], [[1, 1], [1, 1], [1, 1]]
	assert val == [5, 25, 61]

def my_fn_2d(x, y):
	return [x**2 + y**2, x + 2+y]

def test_2d_fn():
	fn = AD(my_fn_2d, ndim=2)
	der = fn.get_der([1,2])
	val = fn.get_val([1,2])
	assert der == [[2, 4], [1, 1]]
	assert val == [5, 5]

def test_func_w_mult_params_single_var():
	fn = lambda x,y: x**2
	ad_fn = AD(fn)
	der = ad_fn.get_der([1,2])
	val = ad_fn.get_val([1,2])
	assert der == [2, 0]
	assert val == 1

def square_fn(x):
        return x ** 2

def test_square_fn():
	ad_square = AD(square_fn)
	der = ad_square.get_der(3)
	val = ad_square.get_val(3)
	assert der == 6
	assert val == 9
	der1 = ad_square.get_der([1,2,3,4])
	val1 = ad_square.get_val([1,2,3,4])
	assert der1 == [2, 4, 6, 8]
	assert val1 == [1, 4, 9, 16]

def test_get_der_types():
	with pytest.raises(TypeError):
        	AD.get_der('string')
	with pytest.raises(TypeError):
        	AD.get_der(dict[1:'a', 2:'b'])

def test_get_val_types():
	with pytest.raises(TypeError):
        	AD.get_val('string')
	with pytest.raises(TypeError):
        	AD.get_val(dict[1:'a', 2:'b'])

def test_get_der_lenlist():
	a = AD(lambda x,y: 3*x**2 + 2*y**3)
	with pytest.raises(Exception):
        	a.get_der(1, 2, 3)
	with pytest.raises(Exception):
			a.get_der([1, 2, 3], [3, 4, 5], [1, 3, 4])

def test_get_val_lenlist():
	a = AD(lambda x,y: 3*x**2 + 2*y**3)
	with pytest.raises(Exception):
        	a.get_val(1, 2, 3)
	with pytest.raises(Exception):
			a.get_val([1, 2, 3], [3, 4, 5], [1, 3, 4])

def test_exception_der():
	fn = AD(my_fn_2d, ndim=2)
	with pytest.raises(Exception):
		fn.get_der([1,2,3])

def test_exception_val():
	fn = AD(my_fn_2d, ndim=2)
	with pytest.raises(Exception):
		fn.get_val([1,2,3])

def my_fn_cos(x):
	return 5*admath.cos(x)

def my_fn_nested_1(x):
	return 5*x**2 * 2*admath.cos(x)

def test_cos():
	cos_fn = AD(my_fn_cos)
	assert cos_fn.get_der(5) == -5*np.sin(5)
	assert cos_fn.get_val(5) == 5*np.cos(5)

def test_nested_1():
	nested_fn = AD(my_fn_nested_1)
	xs = [-10,2,5,10]
	for x in xs:
		assert nested_fn.get_der(x) == pytest.approx(-10*x*(x*np.sin(x)-2*np.cos(x)))
		assert nested_fn.get_val(x) == pytest.approx(5*x**2 * 2*np.cos(x))

def mul_array(x, y):
    return [x**2 * y**2, x + y, 2*x]

def test_mul_array():
	ad_mul_array = AD(mul_array, ndim=3)
	val = ad_mul_array.get_val([[1,2],[1,2],[1,2]])
	der = ad_mul_array.get_der([[1,2],[1,2],[1,2]])
	assert val == [[4, 3, 2], [4, 3, 2], [4, 3, 2]]
	assert der == [[[8, 4], [1, 1], [2, 0]], [[8, 4], [1, 1], [2, 0]], [[8, 4], [1, 1], [2, 0]]]

def functionxx(x):
	return 2*x

def test_numpy_wrapper():
	functionxx_class = AD(functionxx)
	val = functionxx_class.get_val_numpy(np.array([1,2]))
	der = functionxx_class.get_der_numpy(np.array([1,2]))
	assert list(val) == [2,4]
	assert list(der) == [2,2]



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pyautodiff",
    version="0.3.1",
    author="Phoebe Wong",
    author_email="wong@g.harvard.edu",
    description="A package for automatic differentiation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cs207-group2/cs207-FinalProject",
    packages=find_packages('pyautodiff'),
    package_dir = {"":"pyautodiff"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import pytest
from pyautodiff.optimizer import Optimizer

import numpy as np

def test_get_loss2():
    opt1 = Optimizer(loss='mse',optimizer='gd',regularizer='lasso',lam=0.01)
    with pytest.raises(Exception):
        opt1._get_loss_fn('yo')

def test_get_loss3():
    with pytest.raises(Exception):
        opt2=Optimizer(loss='mse',optimizer='homie',regularizer='lasso',lam=0.01)
        opt2._get_loss_fn('homie')

def test_cost():
    opt1 = Optimizer(loss='mse',optimizer='gd',regularizer='lasso',lam=0.01)
    assert type(opt1._get_cost_fn('mse', optimizer='gd')).__name__ == "method"

X2 = np.array([[1,0,1],[1,1,2],[1,2,-2]])
y2 = [-2,-3,21]

def test_opty1():
    opt=Optimizer(loss='mse',optimizer='gd',regularizer='lasso',lam=0.01)
    opt.fit(X2,y2,iters=1000)
    assert opt.coefs[0] <= 3

def user_mse(y,y_preds):
    return (y-y_preds)**2

def test_opty2():
    opt=Optimizer(loss=user_mse,optimizer='sgd',regularizer='ridge',lam=0.01)
    opt.fit(X2,y2,iters=1000)
    assert opt.coefs[0] <= 3

def test_opty3():
    opt=Optimizer(loss=user_mse,optimizer='sgd',lam=0.01)
    opt.fit(X2,y2,iters=1000)
    assert opt.coefs[0] <= 3

def test_opty4():
    opt=Optimizer(loss=user_mse,optimizer='gd',lam=0.01)
    opt.fit(X2,y2,iters=1000)
    assert opt.coefs[0] <= 3

def test_opty_verbose_gd():
    opt=Optimizer(loss=user_mse,optimizer='gd',lam=0.01)
    opt.fit(X2,y2,iters=1000, verbose=True)
    assert opt.coefs[0] <= 3

def test_opty_verbose_sgd():
    opt=Optimizer(loss=user_mse,optimizer='sgd',lam=0.01)
    opt.fit(X2,y2,iters=1000, verbose=True)
    assert opt.coefs[0] <= 3

def test_opty_verbose_steepestgd():
    opt=Optimizer(loss='mse',optimizer='steepestgd',lam=0.01)
    opt.fit(X2,y2,iters=1000, verbose=True)
    assert opt.coefs[0] <= 3

def test_opty_verbose_bfgs():
    opt=Optimizer(loss='mse',optimizer='bfgs',regularizer='ridge',lam=0.01)
    opt.fit(X2,y2,iters=1000, verbose=True)
    assert opt.coefs[0] <= 3

def test_opty_classification():
    opt=Optimizer(loss='mse',optimizer='bfgs',regularizer='ridge',lam=0.01,
                  problem_type='classification')
    opt.fit(X2,y2,iters=1000, verbose=True)
    assert opt.coefs[0] <= 3
    
# def test_opty_regression():
#     opt=Optimizer(loss='mse',optimizer='bfgs',regularizer='ridge',lam=0.01,
#                   problem_type='regression')
#     opt.fit(X2,y2,iters=1000, verbose=True)
#     assert opt.coefs[0] <= 3
    
def test_score():
    opt=Optimizer(loss='mse',optimizer='bfgs',regularizer='ridge',lam=0.01,
                  problem_type='classification')
    opt.fit(X2,y2,iters=1000, verbose=True)
    a = opt.score(X2,y2) 
    assert a>=0 
    
def test_score1():
    opt=Optimizer(loss='mse',optimizer='sgd',regularizer='ridge',lam=0.01,
                  problem_type='classification')
    opt.fit(X2,y2,iters=1000, verbose=True)
    a = opt.score(X2,y2) 
    assert a>=0

def test_score2():
    opt=Optimizer(loss='mse',optimizer='sgd',regularizer='ridge',lam=0.01,
                  problem_type='regression')
    opt.fit(X2,y2,iters=1000, verbose=True)
    a = opt.score(X2,y2) 
    assert a>=0
    
def test_score():
    opt=Optimizer(loss='mse',optimizer='bfgs',regularizer='ridge',lam=0.01,
                  problem_type='classification')
    opt.fit(X2,y2,iters=1000, verbose=True)
    a = opt.predict_proba(X2)[0] 
    assert a>=0 
    